/***************************************************************************

  file.c

  The file management routines

  (c) 2000-2004 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __GB_FILE_C

#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <string.h>

#include "gb_common.h"
#include "gb_error.h"
#include "gb_limit.h"
#include "gb_str.h"

#include "gb_file.h"

PRIVATE char file_buffer[MAX_PATH + 1];
PRIVATE int file_buffer_length;
PUBLIC char *FILE_buffer = file_buffer;

PUBLIC char *FILE_cat(const char *path, ...)
{
  char *p;
  va_list args;
  int len;
  boolean end_slash = FALSE;
  boolean add_slash = FALSE;

  p = file_buffer;
  if (p != path)
    *p = 0;
  va_start(args, path);

  for(;;)
  {
    if (*path == '/' && p != file_buffer)
      path++;

    len = strlen(path);
    if (add_slash)
      len++;

    if (len > 0)
    {
      if ((p + len) > &file_buffer[MAX_PATH])
        return NULL;

      if (add_slash)
        p = STR_add(p, "/");

      if (p == path)
        p += len;
      else
        p = STR_add(p, path);

      end_slash = (p[-1] == '/');
    }

    path = va_arg(args, char *);
    if (path == NULL)
      break;

    add_slash = ((!end_slash) && (*path != 0) && (*path != '/'));
  }

  file_buffer_length = p - file_buffer;
  return file_buffer;
}


PUBLIC int FILE_buffer_length(void)
{
  if (file_buffer_length < 0)
    file_buffer_length = strlen(file_buffer);

  return file_buffer_length;
}


PUBLIC char *FILE_get_dir(const char *path)
{
  char *p;

  if (file_buffer != path)
    strcpy(file_buffer, path);

  p = rindex(file_buffer, '/');

  if (p == NULL)
    *file_buffer = 0;
  else
    *p = 0;

  file_buffer_length = -1;
  return file_buffer;
}


PUBLIC char *FILE_get_ext(const char *path)
{
  char *p;

  p = rindex(path, '/');
  if (p)
    path = &p[1];

  p = rindex(path, '.');
  if (p == NULL)
    return (char *)&path[strlen(path)];
  else
    return p;
}

PUBLIC char *FILE_set_ext(const char *path, const char *ext)
{
  char *p;

  if (file_buffer != path)
    strcpy(file_buffer, path);

  p = FILE_get_ext(file_buffer);
  if (ext == NULL || *ext == 0)
    *p = 0;
  else
  {
    if (*ext != '.')
      *p++ = '.';
    strcpy(p, ext);
  }

  file_buffer_length = -1;
  return file_buffer;
}


PUBLIC time_t FILE_get_time(const char *path)
{
  struct stat info;

  if (stat(path, &info) == 0)
    return info.st_mtime;
  else
    return (time_t)-1L;
}


PUBLIC char *FILE_get_current_dir(void)
{
  return getcwd(file_buffer, MAX_PATH);
}


PUBLIC boolean FILE_exist(const char *path)
{
  struct stat info;

  return stat(path, &info) == 0;
}


PUBLIC char *FILE_get_name(const char *path)
{
  char *p;

  p = rindex(path, '/');
  if (p)
    return &p[1];
  else
    return (char *)path;
}


PUBLIC const char *FILE_get_basename(const char *path)
{
  char *p;

  path = FILE_get_name(path);

  if (file_buffer != path)
    strcpy(file_buffer, path);

  p = rindex(file_buffer, '.');
  if (p)
    *p = 0;

  file_buffer_length = -1;

  return file_buffer;
}


PUBLIC char *FILE_readlink(const char *link)
{
  int len = readlink(link, file_buffer, MAX_PATH);

  if (len < 0)
    return NULL;

  file_buffer[len] = 0;
  file_buffer_length = len;
  return file_buffer;

}
