/***************************************************************************

  CDatabase.c

  The Database class

  (c) 2000-2003 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __CDATABASE_C

#include "main.h"

#include "CTable.h"
#include "CDatabase.h"


/***************************************************************************

  Database

***************************************************************************/

static bool check_database(CDATABASE *_object, char *name, bool must_exist)
{
  bool exist = THIS->driver->Database.Exist(THIS->handle, name);

  if (must_exist)
  {
    if (!exist)
    {
      GB.Error("Unknown database: &1", name);
      return TRUE;
    }
  }
  else
  {
    if (exist)
    {
      GB.Error("Database already exists: &1", name);
      return TRUE;
    }
  }

  return FALSE;
}

static void free_database(CDATABASE* _object)
{
  GB.FreeString(&THIS->database);
}

static void set_database(CDATABASE *_object, char *name)
{
  free_database(THIS);
  GB.NewString(&THIS->database, name, 0);
}


BEGIN_METHOD(CDATABASE_create, GB_STRING name)

  char *name = GB.ToZeroString(ARG(name));

  if (DB_CheckName(name, "database"))
    return;

  if (check_database(THIS, name, FALSE))
    return;

  THIS->driver->Database.Create(THIS->handle, name);

END_METHOD


BEGIN_METHOD(CDATABASE_delete, GB_STRING name)

  char *name = GB.ToZeroString(ARG(name));

  if (check_database(THIS, name, TRUE))
    return;

  THIS->driver->Database.Delete(THIS->handle, name);

END_METHOD


BEGIN_METHOD(CDATABASE_exist, GB_STRING name)

  GB.ReturnBoolean(THIS->driver->Database.Exist(THIS->handle, GB.ToZeroString(ARG(name))));

END_METHOD


BEGIN_METHOD(CDATABASE_find, GB_STRING name)

  char *name = GB.ToZeroString(ARG(name));

  if (check_database(THIS, name, TRUE))
    return;

  set_database(THIS, name);

END_METHOD


BEGIN_PROPERTY(CDATABASE_name)

  GB.ReturnString(THIS->database);

END_PROPERTY


BEGIN_PROPERTY(CDATABASE_system)

  GB.ReturnBoolean(THIS->driver->Database.IsSystem(THIS->handle, THIS->database));

END_PROPERTY


BEGIN_METHOD_VOID(CDATABASE_next)

  int *pos = (int *)GB.GetEnum();

  if (*pos == 0)
  {
    DB_FreeStringArray(&THIS->databases);
    THIS->driver->Database.List(THIS->handle, &THIS->databases);
  }

  if (THIS->databases)
  {
    if (*pos < GB.Count(THIS->databases))
    {
      set_database(THIS, THIS->databases[*pos]);
      (*pos)++;
      return;
    }
  }

  GB.StopEnum();

END_METHOD


GB_DESC CConnectionDatabaseDesc[] =
{
  GB_DECLARE(".ConnectionDatabase", 0), GB_VIRTUAL_CLASS(),

  GB_METHOD("_call", NULL, CDATABASE_find, "(Name)s"),
  GB_METHOD("Find", NULL, CDATABASE_find, "(Name)s"),
  GB_METHOD("Create", NULL, CDATABASE_create, "(Name)s"),
  GB_METHOD("Delete", NULL, CDATABASE_delete, "(Name)s"),
  GB_METHOD("Remove", NULL, CDATABASE_delete, "(Name)s"),
  GB_METHOD("Exist", "b", CDATABASE_exist, "(Name)s"),

  GB_PROPERTY_READ("Name", "s", CDATABASE_name),
  GB_PROPERTY_READ("System", "b", CDATABASE_system),

  GB_METHOD("_next", NULL, CDATABASE_next, NULL),

  GB_END_DECLARE
};



/***************************************************************************

  User

***************************************************************************/

static bool check_user(CDATABASE *_object, char *name, bool must_exist)
{
  bool exist = THIS->driver->User.Exist(THIS->handle, name);

  if (must_exist)
  {
    if (!exist)
    {
      GB.Error("Unknown user: &1", name);
      return TRUE;
    }
  }
  else
  {
    if (exist)
    {
      GB.Error("User already exists: &1", name);
      return TRUE;
    }
  }

  return FALSE;
}

static void free_user(CDATABASE* _object)
{
  GB.FreeString(&THIS->user);
  GB.FreeString(&THIS->user_info.password);
  THIS->user_info.admin = FALSE;
}

static void set_user(CDATABASE *_object, char *name)
{
  free_user(THIS);
  GB.NewString(&THIS->user, name, 0);

  THIS->driver->User.Info(THIS->handle, name, &THIS->user_info);
}


BEGIN_METHOD(CUSER_create, GB_STRING name; GB_STRING password; GB_BOOLEAN admin)

  char *name = GB.ToZeroString(ARG(name));
  DB_USER info;

  CLEAR(&info);

  if (DB_CheckName(name, "user"))
    return;

  if (check_user(THIS, name, FALSE))
    return;

  info.admin = VARGOPT(admin, FALSE);
  if (!MISSING(password))
    info.password = GB.ToZeroString(ARG(password));

  THIS->driver->User.Create(THIS->handle, name, &info);

END_METHOD


BEGIN_METHOD(CUSER_delete, GB_STRING name)

  char *name = GB.ToZeroString(ARG(name));

  if (check_user(THIS, name, TRUE))
    return;

  THIS->driver->User.Delete(THIS->handle, name);

END_METHOD


BEGIN_METHOD(CUSER_exist, GB_STRING name)

  GB.ReturnBoolean(THIS->driver->User.Exist(THIS->handle, GB.ToZeroString(ARG(name))));

END_METHOD


BEGIN_METHOD(CUSER_find, GB_STRING name)

  char *name = GB.ToZeroString(ARG(name));

  if (check_user(THIS, name, TRUE))
    return;

  set_user(THIS, name);

END_METHOD


BEGIN_PROPERTY(CUSER_name)

  GB.ReturnString(THIS->user);

END_PROPERTY


BEGIN_PROPERTY(CUSER_password)

  if (READ_PROPERTY)
    GB.ReturnString(THIS->user_info.password);
  else if (THIS->user)
  {
    GB.StoreString(PROP(GB_STRING), &THIS->user_info.password);
    THIS->driver->User.SetPassword(THIS->handle, THIS->user, THIS->user_info.password);
  }

END_PROPERTY


BEGIN_PROPERTY(CUSER_administrator)

  GB.ReturnBoolean(THIS->user_info.admin);

END_PROPERTY


BEGIN_METHOD_VOID(CUSER_next)

  int *pos = (int *)GB.GetEnum();

  if (*pos == 0)
  {
    DB_FreeStringArray(&THIS->users);
    THIS->driver->User.List(THIS->handle, &THIS->users);
  }

  if (THIS->users)
  {
    if (*pos < GB.Count(THIS->users))
    {
      set_user(THIS, THIS->users[*pos]);
      (*pos)++;
      return;
    }
  }

  GB.StopEnum();

END_METHOD


GB_DESC CConnectionUserDesc[] =
{
  GB_DECLARE(".ConnectionUser", 0), GB_VIRTUAL_CLASS(),

  GB_METHOD("_call", NULL, CUSER_find, "(Name)s"),
  GB_METHOD("Find", NULL, CUSER_find, "(Name)s"),
  GB_METHOD("Create", NULL, CUSER_create, "(Name)s[(Password)s(Administrator)b]"),
  GB_METHOD("Delete", NULL, CUSER_delete, "(Name)s"),
  GB_METHOD("Remove", NULL, CUSER_delete, "(Name)s"),
  GB_METHOD("Exist", "b", CUSER_exist, "(Name)s"),

  GB_PROPERTY_READ("Name", "s", CUSER_name),
  GB_PROPERTY_READ("Administrator", "b", CUSER_administrator),
  GB_PROPERTY("Password", "s", CUSER_password),

  GB_METHOD("_next", NULL, CUSER_next, NULL),

  GB_END_DECLARE
};




/***************************************************************************

  .ConnectionTables

***************************************************************************/

static bool check_table(CDATABASE *_object, char *table, bool must_exist)
{
  bool exist = THIS->driver->Table.Exist(THIS->handle, table);

  if (must_exist)
  {
    if (!exist)
    {
      GB.Error("Unknown table: &1", table);
      return TRUE;
    }
  }
  else
  {
    if (exist)
    {
      GB.Error("Table already exists: &1", table);
      return TRUE;
    }
  }

  return FALSE;
}

static void free_one_dict(CTABLE *table)
{
  GB.Unref((void **)&table);
}


static void free_tables(CDATABASE *_object)
{
  DB_FreeStringArray(&THIS->tables);

  if (THIS->dict)
  {
    GB.Ref(THIS);
    GB.HashTable.Enum(THIS->dict, (GB_HASHTABLE_ENUM_FUNC)free_one_dict);
    GB.HashTable.Free(&THIS->dict);
    GB.Unref((void **)&_object);
  }
}


static CTABLE *get_table(CDATABASE *_object, char *name, bool must_exist)
{
  CTABLE *table;

  if (!THIS->dict)
    GB.HashTable.New(&THIS->dict, GB_COMP_BINARY);

  if (GB.HashTable.Get(THIS->dict, name, 0, (void **)&table))
  {
    if (check_table(THIS, name, must_exist))
      return NULL;

    GB.New((void **)&table, GB.FindClass("Table"), NULL, NULL);

    table->db = THIS;
    table->driver = THIS->driver;
    GB.NewString(&table->table, name, 0);
    table->create = FALSE;

    GB.HashTable.Add(THIS->dict, name, 0, table);
    GB.Ref(table);
    GB.Ref(THIS);
  }

  return table;
}


BEGIN_METHOD(CTABLE_get, GB_STRING name)

  GB.ReturnObject(get_table(THIS, GB.ToZeroString(ARG(name)), TRUE));

END_METHOD


BEGIN_METHOD_VOID(CTABLE_next)

  int *pos = (int *)GB.GetEnum();

  if (*pos == 0)
  {
    DB_FreeStringArray(&THIS->tables);
    THIS->driver->Table.List(THIS->handle, &THIS->tables);
  }

  if (THIS->tables)
  {
    if (*pos < GB.Count(THIS->tables))
    {
      GB.ReturnObject(get_table(THIS, THIS->tables[*pos], TRUE));
      (*pos)++;
      return;
    }
  }

  GB.StopEnum();

END_METHOD


BEGIN_METHOD(CTABLE_create, GB_STRING name)

  CTABLE *table;
  char *name = GB.ToZeroString(ARG(name));

  if (DB_CheckName(name, "table"))
    return;

  table = get_table(THIS, name, FALSE);

  if (table)
  {
    table->create = TRUE;
    GB.ReturnObject(table);
  }

END_METHOD


BEGIN_METHOD(CTABLE_delete, GB_STRING name)

  char *table = GB.ToZeroString(ARG(name));

  if (check_table(THIS, table, TRUE))
    return;

  THIS->driver->Table.Delete(THIS->handle, table);

END_METHOD


BEGIN_METHOD(CTABLE_exist, GB_STRING name)

  GB.ReturnBoolean(THIS->driver->Table.Exist(THIS->handle, GB.ToZeroString(ARG(name))));

END_METHOD


BEGIN_METHOD_VOID(CTABLE_refresh)

  free_tables(THIS);

END_METHOD



GB_DESC CConnectionTablesDesc[] =
{
  GB_DECLARE(".ConnectionTables", 0), GB_VIRTUAL_CLASS(),

  GB_METHOD("_get", "Table", CTABLE_get, "(Name)s"),
  GB_METHOD("_next", "Table", CTABLE_next, NULL),

  GB_METHOD("Refresh", NULL, CTABLE_refresh, NULL),

  GB_METHOD("Create", "Table", CTABLE_create, "(Name)s"),
  GB_METHOD("Delete", NULL, CTABLE_delete, "(Name)s"),
  GB_METHOD("Remove", NULL, CTABLE_delete, "(Name)s"),
  GB_METHOD("Exist", "b", CTABLE_exist, "(Name)s"),

  GB_END_DECLARE
};



/***************************************************************************

  Connection

***************************************************************************/

static CDATABASE *_current = NULL;


static bool check_opened(CDATABASE *_object)
{
  if (!THIS->handle)
  {
    GB.Error("Database is not opened");
    return TRUE;
  }
  else
    return FALSE;
}

#define CHECK_OPEN() \
  if (check_opened(THIS)) \
    return;

static int get_current(CDATABASE **current)
{
  if (*current == NULL)
  {
    if (_current == NULL)
    {
      GB.Error("No current database");
      return TRUE;
    }
    *current = _current;
  }

  return FALSE;
}

#define CHECK_DB() \
  if (get_current((CDATABASE **)&_object)) \
    return;

static void close_database(CDATABASE *_object)
{
  if (THIS->handle)
  {
    free_tables(THIS);
    DB_FreeStringArray(&THIS->databases);
    DB_FreeStringArray(&THIS->users);
    free_database(THIS);
    free_user(THIS);

    THIS->driver->Close(THIS->handle);
    THIS->handle = NULL;

    if (_current == THIS)
      _current = NULL;
  }
}

// BEGIN_METHOD(CDATABASE_new, GB_STRING url)
BEGIN_METHOD_VOID(CDATABASE_new)

  /*int i, state, p;
  char c;*/

  THIS->handle = NULL;
  THIS->dict = NULL;

  if (_current == NULL)
    _current = THIS;

#if 0
  if (!MISSING(url))
  {
    char *url = GB.ToZeroString(ARG(url));

    state = 0;
    p = 0;
    for (i = 0; i <= LENGTH(url); i++)
    {
      c = url[i];

      switch (state)
      {
        case 0: /* type */
          if (c == ':' || c == 0)
          {
            GB.NewString(&THIS->desc.type, &url[p], i - p);
            state++;
            p = i + 3; /* On saute '://' */
          }
          break;

        case 1: /* // */
        case 2:
          if (c && c != '/')
            goto BAD_URL;

          state++;
          break;

        case 3: /* host */
          if (c == ':' || c == '/' || c == 0)
          {
            if (i > p)
              GB.NewString(&THIS->desc.host, &url[p], i - p);
            p = i + 1;
            state++;
            if (c != ':')
              state++;
          }
          break;

        case 4: /* port */
          if (c == '/' || c == 0)
          {
            if (i == p)
              goto BAD_URL;

            GB.NewString(&THIS->desc.port, &url[p], i - p);
            p = i + 1;
            state++;
          }
          break;

        case 5: /* name */
          if (c == 0 && i > p)
            GB.NewString(&THIS->desc.name, &url[p], i - p);
          break;
      }
    }
  }

  return;

BAD_URL:

  GB.Error("Malformed URL");
  return;
#endif

END_METHOD


BEGIN_METHOD_VOID(CDATABASE_free)

  close_database(THIS);

  #if 0
  if (THIS->tables)
    GB.Unref((void **)&THIS->tables);
  #endif

  GB.StoreString(NULL, &THIS->desc.type);
  GB.StoreString(NULL, &THIS->desc.host);
  GB.StoreString(NULL, &THIS->desc.user);
  GB.StoreString(NULL, &THIS->desc.password);
  GB.StoreString(NULL, &THIS->desc.name);
  GB.StoreString(NULL, &THIS->desc.port);

  GB.StoreString(NULL, &THIS->charset);

  #if 0
  free_string_array(&THIS->databases);
  free_string_array(&THIS->users);
  #endif

END_METHOD


#define IMPLEMENT(_prop) \
BEGIN_PROPERTY(CCONNECTION_##_prop) \
\
  if (READ_PROPERTY) \
    GB.ReturnString(THIS->desc._prop); \
  else \
    GB.StoreString(PROP(GB_STRING), &THIS->desc._prop); \
\
END_PROPERTY

IMPLEMENT(type)
IMPLEMENT(host)
IMPLEMENT(user)
IMPLEMENT(password)
IMPLEMENT(name)
IMPLEMENT(port)


BEGIN_METHOD_VOID(CDATABASE_open)

  CHECK_DB();

  if (THIS->handle)
  {
    GB.Error("Database already opened.");
    return;
  }

  DB_Open(&THIS->desc, &THIS->driver, &THIS->handle, &THIS->charset);

END_METHOD


BEGIN_METHOD_VOID(CDATABASE_close)

  CHECK_DB();

  close_database(THIS);

END_METHOD


BEGIN_METHOD_VOID(CDATABASE_begin)

  CHECK_DB();

  CHECK_OPEN();

  THIS->driver->Begin(THIS->handle);

END_METHOD


BEGIN_METHOD_VOID(CDATABASE_commit)

  CHECK_DB();

  CHECK_OPEN();

  THIS->driver->Commit(THIS->handle);

END_METHOD


BEGIN_METHOD_VOID(CDATABASE_rollback)

  CHECK_DB();

  CHECK_OPEN();

  THIS->driver->Rollback(THIS->handle);

END_METHOD


BEGIN_METHOD(CCONNECTION_exec, GB_STRING query; GB_VALUE param[0])

  char *query;
  CRESULT *result;

  CHECK_DB();

  CHECK_OPEN();

  query = DB_MakeQuery(THIS->driver, STRING(query), LENGTH(query), GB.NParam(), ARG(param[0]));
  if (!query)
    return;

  result = DB_MakeResult(THIS, RESULT_FIND, NULL, query);

  if (result)
    GB.ReturnObject(result);

END_METHOD


BEGIN_METHOD(CCONNECTION_create, GB_STRING table)

  CRESULT *result;
  char *table = GB.ToZeroString(ARG(table));

  CHECK_DB();

  CHECK_OPEN();

  /*if (check_table(THIS, table, TRUE))
    return;*/

  result = DB_MakeResult(THIS, RESULT_CREATE, table, NULL);

  if (result)
    GB.ReturnObject(result);

END_METHOD


static char *get_query(CDATABASE *_object, char *table, long len_table, char *query, long len_query, GB_VALUE *arg)
{
  q_init();

  q_add("SELECT * FROM ");
  q_add_length(table, len_table);

  if (query)
  {
    q_add(" WHERE ");
    q_add_length(query, len_query);

    query = DB_MakeQuery(THIS->driver, q_get(), q_length(), GB.NParam(), arg);
  }
  else
  {
    query = q_get();
  }

  return query;
}


BEGIN_METHOD(CCONNECTION_find, GB_STRING table; GB_STRING query; GB_VALUE param[0])

  char *query;
  CRESULT *result;
  /*char *table = GB.ToZeroString(ARG(table));*/

  CHECK_DB();

  CHECK_OPEN();

  /*if (check_table(THIS, table, TRUE))
    return;*/

  query = get_query(THIS, STRING(table), LENGTH(table),
    MISSING(query) ? NULL : STRING(query),
    MISSING(query) ? 0 : LENGTH(query),
    ARG(param[0]));

  if (!query)
    return;

  result = DB_MakeResult(THIS, RESULT_FIND, NULL, query);

  if (result)
    GB.ReturnObject(result);

END_METHOD


BEGIN_METHOD(CCONNECTION_edit, GB_STRING table; GB_STRING query; GB_VALUE param[0])

  char *query;
  CRESULT *result;
  /*char *table = GB.ToZeroString(ARG(table));*/

  CHECK_DB();

  CHECK_OPEN();

  /*if (check_table(THIS, table, TRUE))
    return;*/

  query = get_query(THIS, STRING(table), LENGTH(table),
    MISSING(query) ? NULL : STRING(query),
    MISSING(query) ? 0 : LENGTH(query),
    ARG(param[0]));

  if (!query)
    return;

  result = DB_MakeResult(THIS, RESULT_EDIT, GB.ToZeroString(ARG(table)), query);

  if (result)
    GB.ReturnObject(result);

END_METHOD


BEGIN_PROPERTY(CDATABASE_current)

  if (READ_PROPERTY)
    GB.ReturnObject(_current);
  else
    _current = (CDATABASE *)VPROP(GB_OBJECT);

END_PROPERTY


BEGIN_PROPERTY(CCONNECTION_self)

  CHECK_DB();
  CHECK_OPEN();
  RETURN_SELF();

END_PROPERTY


BEGIN_PROPERTY(CCONNECTION_charset)

  CHECK_DB();
  CHECK_OPEN();
  if (THIS->charset)
    GB.ReturnString(THIS->charset);
  else
    GB.ReturnConstZeroString("ASCII");

END_PROPERTY



GB_DESC CConnectionDesc[] =
{
  GB_DECLARE("Connection", sizeof(CDATABASE)),

  GB_METHOD("_new", NULL, CDATABASE_new, NULL), //"[(URL)s]"),
  GB_METHOD("_free", NULL, CDATABASE_free, NULL),

  GB_PROPERTY("Type", "s", CCONNECTION_type),
  GB_PROPERTY("Host", "s", CCONNECTION_host),
  GB_PROPERTY("Login", "s", CCONNECTION_user),
  GB_PROPERTY("Password", "s", CCONNECTION_password),
  GB_PROPERTY("Name", "s", CCONNECTION_name),
  GB_PROPERTY("Port", "s", CCONNECTION_port),

  GB_METHOD("Open", NULL, CDATABASE_open, NULL),
  GB_METHOD("Close", NULL, CDATABASE_close, NULL),

  GB_PROPERTY_READ("Charset", "s", CCONNECTION_charset),
  //GB_PROPERTY("

  GB_METHOD("Exec", "Result", CCONNECTION_exec, "(Request)s(Arguments)."),
  GB_METHOD("Create", "Result", CCONNECTION_create, "(Table)s"),
  GB_METHOD("Find", "Result", CCONNECTION_find, "(Table)s[(Request)s(Arguments).]"),
  GB_METHOD("Edit", "Result", CCONNECTION_edit, "(Table)s[(Request)s(Arguments).]"),

  GB_METHOD("Begin", NULL, CDATABASE_begin, NULL),
  GB_METHOD("Commit", NULL, CDATABASE_commit, NULL),
  GB_METHOD("Rollback", NULL, CDATABASE_rollback, NULL),

  GB_PROPERTY("Tables", ".ConnectionTables", CCONNECTION_self),

  GB_PROPERTY("Database", ".ConnectionDatabase", CCONNECTION_self),
  GB_PROPERTY("User", ".ConnectionUser", CCONNECTION_self),

  GB_END_DECLARE
};


GB_DESC CDBDesc[] =
{
  GB_DECLARE("DB", 0), GB_VIRTUAL_CLASS(),

  GB_STATIC_PROPERTY("Current", "Connection", CDATABASE_current),

  GB_STATIC_METHOD("Open", NULL, CDATABASE_open, NULL),
  GB_STATIC_METHOD("Close", NULL, CDATABASE_close, NULL),

  GB_STATIC_PROPERTY_READ("Charset", "s", CCONNECTION_charset),

  GB_STATIC_METHOD("Exec", "Result", CCONNECTION_exec, "(Request)s(Arguments)."),
  GB_STATIC_METHOD("Create", "Result", CCONNECTION_create, "(Table)s"),
  GB_STATIC_METHOD("Find", "Result", CCONNECTION_find, "(Table)s[(Request)s(Arguments).]"),
  GB_STATIC_METHOD("Edit", "Result", CCONNECTION_edit, "(Table)s[(Request)s(Arguments).]"),

  GB_STATIC_METHOD("Begin", NULL, CDATABASE_begin, NULL),
  GB_STATIC_METHOD("Commit", NULL, CDATABASE_commit, NULL),
  GB_STATIC_METHOD("Rollback", NULL, CDATABASE_rollback, NULL),

  GB_STATIC_PROPERTY("Tables", ".ConnectionTables", CCONNECTION_self),

  GB_STATIC_PROPERTY("Database", ".ConnectionDatabase", CCONNECTION_self),
  GB_STATIC_PROPERTY("User", ".ConnectionUser", CCONNECTION_self),

  GB_END_DECLARE
};


